package javax.xml.xpath;

import java.net.URL;
import java.util.ServiceConfigurationError;
import java.util.Iterator;
import java.util.ServiceLoader;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.security.AccessControlContext;
import java.lang.reflect.InvocationTargetException;
import com.sun.org.apache.xpath.internal.jaxp.XPathFactoryImpl;
import java.io.File;
import java.util.Properties;

class XPathFactoryFinder
{
    private static final String DEFAULT_PACKAGE = "com.sun.org.apache.xpath.internal";
    private static final SecuritySupport ss;
    private static boolean debug;
    private static final Properties cacheProps;
    private static volatile boolean firstTime;
    private final ClassLoader classLoader;
    private static final Class<XPathFactory> SERVICE_CLASS;
    
    private static void debugPrintln(String msg) {
        if (XPathFactoryFinder.debug) {
            System.err.println("JAXP: " + msg);
        }
    }
    
    public XPathFactoryFinder(ClassLoader loader) {
        this.classLoader = loader;
        if (XPathFactoryFinder.debug) {
            this.debugDisplayClassLoader();
        }
    }
    
    private void debugDisplayClassLoader() {
        try {
            if (this.classLoader == XPathFactoryFinder.ss.getContextClassLoader()) {
                debugPrintln("using thread context class loader (" + this.classLoader + ") for search");
                return;
            }
        }
        catch (Throwable t) {}
        if (this.classLoader == ClassLoader.getSystemClassLoader()) {
            debugPrintln("using system class loader (" + this.classLoader + ") for search");
            return;
        }
        debugPrintln("using class loader (" + this.classLoader + ") for search");
        return;
    }
    
    public XPathFactory newFactory(String uri) throws XPathFactoryConfigurationException {
        if (uri == null) {
            throw new NullPointerException();
        }
        XPathFactory f = this._newFactory(uri);
        if (f != null) {
            debugPrintln("factory '" + f.getClass().getName() + "' was found for " + uri);
            return f;
        }
        debugPrintln("unable to find a factory for " + uri);
        return f;
    }
    
    private XPathFactory _newFactory(String uri) throws XPathFactoryConfigurationException {
        XPathFactory xpathFactory = null;
        String propertyName = XPathFactoryFinder.SERVICE_CLASS.getName() + ":" + uri;
        try {
            debugPrintln("Looking up system property '" + propertyName + "'");
            String r = XPathFactoryFinder.ss.getSystemProperty(propertyName);
            if (r != null) {
                debugPrintln("The value is '" + r + "'");
                xpathFactory = this.createInstance(r);
                if (xpathFactory != null) {
                    return xpathFactory;
                }
            }
            else {
                debugPrintln("The property is undefined.");
            }
        }
        catch (Throwable t) {
            if (XPathFactoryFinder.debug) {
                debugPrintln("failed to look up system property '" + propertyName + "'");
                t.printStackTrace();
            }
        }
        String javah = XPathFactoryFinder.ss.getSystemProperty("java.home");
        String configFile = javah + File.separator + "lib" + File.separator + "jaxp.properties";
        try {
            if (XPathFactoryFinder.firstTime) {
                synchronized (XPathFactoryFinder.cacheProps) {
                    if (XPathFactoryFinder.firstTime) {
                        File f = new File(configFile);
                        XPathFactoryFinder.firstTime = false;
                        if (XPathFactoryFinder.ss.doesFileExist(f)) {
                            debugPrintln("Read properties file " + f);
                            XPathFactoryFinder.cacheProps.load(XPathFactoryFinder.ss.getFileInputStream(f));
                        }
                    }
                }
            }
            String factoryClassName = XPathFactoryFinder.cacheProps.getProperty(propertyName);
            debugPrintln("found " + factoryClassName + " in $java.home/jaxp.properties");
            if (factoryClassName != null) {
                xpathFactory = this.createInstance(factoryClassName);
                if (xpathFactory != null) {
                    return xpathFactory;
                }
            }
        }
        catch (Exception ex) {
            if (XPathFactoryFinder.debug) {
                ex.printStackTrace();
            }
        }
        assert xpathFactory == null;
        xpathFactory = this.findServiceProvider(uri);
        if (xpathFactory != null) {
            return xpathFactory;
        }
        if (uri.equals("http://java.sun.com/jaxp/xpath/dom")) {
            debugPrintln("attempting to use the platform default W3C DOM XPath lib");
            return new XPathFactoryImpl();
        }
        debugPrintln("all things were tried, but none was found. bailing out.");
        return null;
    }
    
    private Class<?> createClass(String className) {
        boolean internal = false;
        if (System.getSecurityManager() != null && className != null && className.startsWith("com.sun.org.apache.xpath.internal")) {
            internal = true;
        }
        Class<Object> clazz;
        try {
            if (this.classLoader != null && !internal) {
                clazz = Class.forName(className, false, this.classLoader);
            }
            else {
                clazz = Class.forName(className);
            }
        }
        catch (Throwable t) {
            if (XPathFactoryFinder.debug) {
                t.printStackTrace();
            }
            return null;
        }
        return clazz;
    }
    
    XPathFactory createInstance(String className) throws XPathFactoryConfigurationException {
        XPathFactory xPathFactory = null;
        debugPrintln("createInstance(" + className + ")");
        Class<?> clazz = this.createClass(className);
        if (clazz == null) {
            debugPrintln("failed to getClass(" + className + ")");
            return null;
        }
        debugPrintln("loaded " + className + " from " + which(clazz));
        try {
            xPathFactory = (XPathFactory)clazz.getConstructor(new Class[0]).newInstance(new Object[0]);
        }
        catch (ClassCastException | IllegalAccessException | IllegalArgumentException | InstantiationException | InvocationTargetException | NoSuchMethodException | SecurityException ex) {
            debugPrintln("could not instantiate " + clazz.getName());
            if (XPathFactoryFinder.debug) {
                ex.printStackTrace();
            }
            return null;
        }
        return xPathFactory;
    }
    
    private boolean isObjectModelSupportedBy(final XPathFactory factory, final String objectModel, AccessControlContext acc) {
        return (boolean)AccessController.doPrivileged(new PrivilegedAction<Boolean>() {
            @Override
            public Boolean run() {
                return Boolean.valueOf(factory.isObjectModelSupported(objectModel));
            }
        }, acc);
    }
    
    private XPathFactory findServiceProvider(final String objectModel) throws XPathFactoryConfigurationException {
        assert objectModel != null;
        final AccessControlContext acc = AccessController.getContext();
        try {
            return (XPathFactory)AccessController.doPrivileged(new PrivilegedAction<XPathFactory>() {
                @Override
                public XPathFactory run() {
                    ServiceLoader<XPathFactory> loader = ServiceLoader.load(XPathFactoryFinder.SERVICE_CLASS);
                    for (XPathFactory factory : loader) {
                        if (XPathFactoryFinder.access$100(XPathFactoryFinder.this, factory, objectModel, acc)) {
                            return factory;
                        }
                        continue;
                    }
                    return null;
                }
            });
        }
        catch (ServiceConfigurationError error) {
            throw new XPathFactoryConfigurationException(error);
        }
    }
    
    private static String which(Class clazz) {
        return which(clazz.getName(), clazz.getClassLoader());
    }
    
    private static String which(String classname, ClassLoader loader) {
        String classnameAsResource = classname.replace('.', '/') + ".class";
        if (loader == null) {
            loader = ClassLoader.getSystemClassLoader();
        }
        URL it = XPathFactoryFinder.ss.getResourceAsURL(loader, classnameAsResource);
        if (it != null) {
            return it.toString();
        }
        return null;
    }
    
    static boolean access$100(XPathFactoryFinder x0, XPathFactory x1, String x2, AccessControlContext x3) {
        return x0.isObjectModelSupportedBy(x1, x2, x3);
    }
    
    static {
        ss = new SecuritySupport();
        XPathFactoryFinder.debug = false;
        try {
            XPathFactoryFinder.debug = (XPathFactoryFinder.ss.getSystemProperty("jaxp.debug") != null);
        }
        catch (Exception unused) {
            XPathFactoryFinder.debug = false;
        }
        cacheProps = new Properties();
        XPathFactoryFinder.firstTime = true;
        SERVICE_CLASS = XPathFactory.class;
    }
}
